/* bottomlevel.c - chunk-level base primitiaves for the meme machine
 * 
 * Primitives in this file:
 * read_text()
 * get_chunk()
 * free_chunk()
 * dump_stats()
 * load()
 * save()
 */
 
#include <stdio.h>
#include <stdlib.h>
#include <sys/mman.h>
#include "datastru.h"

/***************************************************************
 * read_text(input);
 * This routine reads in the input text file to memory.
 * The routine also serves to filter out all the crap characters
 * in the input text stream.
 */
 
void read_text(net_type *net, nodelist_type *input, nodelist_type *output,
	       char *filename)
{
  FILE *fp_in; int filesize;
  char cha;
  int loop;
  int quote;    /*boolean - whether we are within a quote or not */
  int line, qline;  /*current line being processed... */
  
  input->size=0;  quote=0; line=0; qline=0;
  
  if (!(fp_in=fopen(filename,"r")))
  {
     fprintf(stderr,"training file '%s'  not found - bummer: ",filename);
     perror("fopen");
     exit(1);
  }
  /* firstly, see just how big this training file is... */
  fseek(fp_in,0,2); filesize=ftell(fp_in); fseek(fp_in,0,0);
  input->data=(int *)malloc(filesize*sizeof(int));
  output->data=(int *)malloc(filesize*2*sizeof(int));
  while(!feof(fp_in))
  {    
    cha=getc(fp_in);  
    if (net->chatlog && cha==10 && input->size>2 ) cha='.';
    if (cha==10) { line++; cha=32; }
    if (cha==-1) cha=32;
    if (cha>=32)
    {
      input->size++;
      if (cha>='A' && cha<='Z') cha=cha-'A'+'a'; /* get everything lower-case */
      input->data[input->size]=(int)cha;
      if (net->mode==69 && cha==34) {  /*we hit a quote and are in story mode*/
        if (quote==0) {
	  quote=1;
	  input->size--;
	  qline=line;
	}/*begin quote... */
	else 
	{
	   quote=0; input->data[input->size]=10;
	   if (line-qline>MAX_QUOTED_LINES)
	       printf("**** POSSIBLE QUOTES-ERROR AT LINE %i\n",qline);
	}
      }
      /* only include quoted text, with option 69... */
      if (net->mode==69 && cha!=34 && quote==0) input->size--;
    }
    /* bit of extra processing... remove space between terminator chars */
    if (input->size>2 && responce_node(input->data[input->size-1]) &&
        input->data[input->size]==32)
		input->size--;
    /* also, prevent 2 terminator chars sitting next to each other */
    if (input->size>2 && responce_node(input->data[input->size-1]) &&
    			 responce_node(input->data[input->size]))
		input->size--;
    /* prevent 2 spaces sitting next to each other... */
    if (input->size>2 && input->data[input->size-1]==' ' &&
                         input->data[input->size  ]==' ')
	 	input->size--;  
  }
  input->size--;
  /* lastly, we need to multiply all the input by TOP... */
  for (loop=1; loop<=input->size; loop++)
  {
    /* printf("%c",input->data[loop]); */
    input->data[loop]=input->data[loop]*TOP;
  }
  if (net->mode==69) net->mode=1;
}

/**************************************************************
 * int get_chunk(net,size)
 * This routine will return a pointer to an area in net.data[]
 * of 'size' integers. more space at the top of net.data will
 * be provided if necessary, otherwise a previously allocted and
 * freed-up chunk will be recycled.
 * NOTE -  It is assumed that data will automatically
 * be copied into the new space.
 * NOTE - linked-list of free chunks is a First-in-First-out system.
 */

int get_chunk(net_type *net, int size)
{
    int posn;
    int trace, traceposn, tracetotal;
    
    /* firstly, point all requests for 0-sized chunks to NULL */
    if (size==0) return(0);
    
    /* worst-case scenario check - need to chop an existing free chunk up*/
    /* (the routine here is not well tested, so left out for now. */
    /* if ((net->ptr-size)<=net->nodespace && net->free[size]==0)
       chop_a_chunk(net,size); */
    
    /*Firstly, see if a previously allocated area can be reclamed*/
    if (net->free[size]!=0)  /*i.e. a previous freed chunk available*/
    {
      posn=net->free[size];
      net->free[size]=net->data[posn];
    }
    else /*just grab new space at the top of net->data*/
    {      
      net->ptr=net->ptr-size;
      posn=net->ptr;
      if ((net->ptr-size)<=net->nodespace)
      {
         printf("Woa! get_chunk()- net.data over-run allocated size.Bummer.\n");
         printf("increase MAX_DATA_SIZE in datastructs.h, rm *.o, remove \n");
         printf("the net file, and do a make.\n");
	 printf("problem chunk-size=%i\n",size);
	 printf("tracedump of free spaces follows... \n");
	 for (trace=1; trace<MAX_IPOP; trace++)
	 {
	    traceposn=net->free[trace]; tracetotal=0;
	    while(traceposn!=0) 
	       {traceposn=net->data[traceposn]; tracetotal++;}
	    if (tracetotal!=0)
	      printf("size %i, free=%i\n",trace,tracetotal);
	 }
         exit(0);
      }
    }
    return(posn);
}


/*****************************************************************
 * free_chunk(net,start,size)
 * This routine will free up a chunk of net.data at location 'start')
 */
void free_chunk(net_type *net, int start, int size)
{
  /* corner case - ignore requests to free up a 0-sized chunk */
  if (size==0);
  /*put a link into previous net->free[size] in net[start], and change
    net->free[size] to be start */
  if(size>MAX_IPOP)
  {
    /*bollocks, hopefully there wont be a chunk of free space more
     than MAX_IPOP -integers. If there is, write some more code here! */
     printf("WOAAA! free_chunk() - dont like this at all!\n");
     exit(0);
  }
  net->data[start]=net->free[size]; /*pointer to next free chunk */
  net->free[size]=start;  
}


/**********************************************************************
 * chop_a_chunk()
 * This nifty recursive routine will chop a free chunk into two to allow
 * better allocation of space later on...
 * (CURRENTLY UNUSED AND UNTESTED - DEFRAGGING IS PREFERRED OPTION...)
 */

void chop_a_chunk(net_type *net,int size)
{
   int new_chunk;
   
   printf("*** inside chop_a_chunk() - size=%i\n",size);
   /* sanity check... */
   if (size>MAX_IPOP) {printf("WOA! out of space! no luck!!!\n"); exit(0); }
   if (size<1024)
   {
     if (net->free[size*2]!=0)
     {
        new_chunk=get_chunk(net,size*2);
	free_chunk(net,new_chunk,size); free_chunk(net,new_chunk+size,size);
     }
     else
     {
       chop_a_chunk(net,size*2);
       chop_a_chunk(net,size);
       return;
     }
     
   }
   else  /* size>1024, chop 1024 off of a higher chunk... */
   {
      if (net->free[size+1024]!=0)
      {
        new_chunk=get_chunk(net,size+1024);
	free_chunk(net,new_chunk,size); free_chunk(net,new_chunk+size,1024);
      }
      else
      {
         chop_a_chunk(net,size+1024);
	 chop_a_chunk(net,size);
	 return;
      }
   }
   
}

/**********************************************************************
 * dump_stats()
 * This routine will dump out stats on chunk-sizes used in a net.
 */

void dump_stats(net_type *net)
{
   int usedcount[MAX_IPOP];
   int level[MAX_LEVELS];
   int trace, traceposn,tracetotal;
   
  /* firstly, count the number of free items of each slotsize and report*/
  printf("\n----------dump of stats on net------------\n");
  printf("------------------------------------------\n");
  printf("Stats-dump of free spaces follows... \n");
  for (trace=1; trace<MAX_IPOP; trace++)
  {
    traceposn=net->free[trace]; tracetotal=0;
    while(traceposn!=0) 
       {traceposn=net->data[traceposn]; tracetotal++;}
        if (tracetotal!=0)
	     printf("size %i, total=%i\n",trace,tracetotal);
  }
  printf("\n---------------------------------------------\n");
  printf("Stats-dump of Used spaces follows...\n");
  /* now the more complex count of sizes of chunks allocated... */
  for (trace=1; trace<MAX_IPOP; trace++) usedcount[trace]=0;
  for (trace=0; trace<net->nodespace; trace+=TOP)
  {
     usedcount[net->data[trace+NUM_LHS_OP]]++;
     usedcount[net->data[trace+NUM_RHS_OP]]++;
  }
  for (trace=1; trace<MAX_IPOP; trace++)
    if (usedcount[trace]!=0)
      printf("size %i, total=%i\n",trace,usedcount[trace]);
  printf("------------------------------------------\n");
  printf("---- count of nodes at each level... ---- \n");
  for (trace=1; trace<MAX_LEVELS; trace++) level[trace]=0;
  for (trace=0; trace<net->nodespace; trace+=TOP)
     level[get_level(net,trace)]++;
  for (trace=1; trace<MAX_LEVELS; trace++)
    if (level[trace]!=0) printf("level %i, count=%i\n",trace,level[trace]);
  printf("----------- end of stats ----------\n");
}

/* The following 2 routines are for the poor people who have
   MS Windows... */

/*****************************************************************
 * load(net);
 * This routine reads in the construct data from a file
 */

void load(net_type *net)
{
  FILE *fp;

  if (!(fp=fopen("net","rb"))) {
    perror("fopen('net')");
    exit(1);
  }
  fread(&(net->ptr),sizeof(int),1,fp);
  fread(&(net->nodespace),sizeof(int),1,fp);
  fread(&(net->actual_num_nodes),sizeof(int),1,fp);
  fread(&(net->deleted_node),sizeof(int),1,fp);
  fread(&(net->next_dream),sizeof(int),1,fp);
  fread(&(net->free),sizeof(int),MAX_IPOP+1,fp);   
  fread(&(net->data),sizeof(int),MAX_DATA_SIZE,fp);
  if (fclose(fp)) {
    perror("fclose('net')");
    exit(1);
  }
}

/****************************************************************
 * save(net);
 * This routine writes volitile data to disk at the end
 *  of the program
 */

void save(net_type *net)
{
  FILE *fp;

  if (!(fp=fopen("net","wb"))) {
    perror("fopen('net')");
    exit(1);
  }
  fwrite(&(net->ptr),sizeof(int),1,fp);
  fwrite(&(net->nodespace),sizeof(int),1,fp);
  fwrite(&(net->actual_num_nodes),sizeof(int),1,fp);
  fwrite(&(net->deleted_node),sizeof(int),1,fp);
  fwrite(&(net->next_dream),sizeof(int),1,fp);
  fwrite(&(net->free),sizeof(int),MAX_IPOP+1,fp);
  fwrite(&(net->data),sizeof(int),MAX_DATA_SIZE,fp);
  if (fclose(fp)) {
    perror("fclose('net')");
    exit(1);
  }
}                                                     
 
 
