/* midlevel.c - node-creation and deletion
 *
 * Primitives defined here:
 * new_node()
 * make_node()
 * delete_node()
 */


#include <stdio.h>
#include "datastru.h"

/***************************************************************
 * int new_node(net)
 * This routine gives back a node number, the next
 * allocated from net->node or net.free_node_stack. 
 * Initially the node has no inputs or outputs.
 * the 'details' indicator is set to 0.
 * the 'level' indicator is set to 1 (new nodes are by default base nodes).
 * This routine should ONLY EVER be called from make_node()
 */
 
int new_node(net_type *net)
{
   int node,loop;
   
   /* get the next node number from the deleted_node stack if possible */
   if (net->deleted_node!=0)
   {
     node=net->deleted_node;
     net->deleted_node=net->data[node+SPEC];
     net->actual_num_nodes++;
   }
   else
   {
     net->nodespace+=TOP; net->actual_num_nodes++;
     if (net->nodespace>=net->ptr)
     {
       printf("Woa! new_node()- net.data over-run allocated size.Bummer.\n");
       printf("increase MAX_DATA_SIZE in datastructs.h, rm *.o, remove \n");
       printf("the net file, and do a make.\n");
       exit(0);
     }
     node=net->nodespace;
   }
   /* blank out the allocated node.... */
   for (loop=node; loop<node+TOP; loop++) net->data[loop]=0;
   return(node);
}


/**************************************************************************
 * make_node(net, lhs, rhs)
 * create a new node with inputs lhs and rhs (left-hand and right-hand-side)
 * returns the node created. Returns 0 if the node already exists.
 * 
 * called from netlisten()
 */

int make_node(net_type *net, int lhs, int rhs)
{
  int node;
  int found;
  int lhs_level,rhs_level;
  int tracer;

  /* printf("make node, '"); print_node(net,lhs); printf("' and '");
     print_node(net,rhs); printf("'\n"); */
  /* return(0); */ /* This is used to force no-learning during tracing.*/
  /* dont make a node if it's been marked as not to be created */
  if (lhs==0) return(0); if (rhs==0) return(0);
  /* tracer */
  if (lhs>net->nodespace) {printf("WOA!!! lhs\n"); exit(0); }
  if (rhs>net->nodespace) {printf("WOA!!! rhs\n"); exit(0); }
  /* firstly, dont make the node if it already exists!!! */
  found=find_node(net,lhs,rhs); if (found!=0) return(found);
  node=new_node(net);
  net->data[node+LHS_IP]=lhs; net->data[node+RHS_IP]=rhs;
  /* now add appropriate outputs to other nodes... */
  add_lhs_output(net,lhs,node);
  add_rhs_output(net,rhs,node);
  /* set level of node appropriately... */
  lhs_level=(net->data[lhs] & LEVEL_MASK) / LEVEL_DIV;
  rhs_level=(net->data[rhs] & LEVEL_MASK) / LEVEL_DIV;
  if (lhs_level>rhs_level)
  {lhs_level++; net->data[node]+=lhs_level*LEVEL_DIV; }
  else
  {rhs_level++; net->data[node]+=rhs_level*LEVEL_DIV; }
  /* another tracer...\n"); 
     if (get_level(net,node)==0) {printf("WOA! levels screwed!\n"); exit(0); } */
  /* Now see if new node is a responce node or not */
  if (net->data[lhs]>=HIGH_BIT)
          net->data[node]+=HIGH_BIT;
  else 
  if (net->data[rhs]>=HIGH_BIT)
          net->data[node]+=HIGH_BIT;
   
  /* set up net->first for the node...*/
  net->data[node]+=net->data[lhs] & FIRST_MASK;
  tracer=(net->data[node] & FIRST_MASK) / 8; /*first-char */
  /* printf("*** new node - '"); print_node(net,node); 
     printf("', first=%i\n", tracer); */
  if (tracer>127) {printf("WOA! tracer=%i\n",tracer); exit(0); }
  
  return(node);
}

/*************************************************************************
 * delete_node(net,node)
 * delete a node and put it in the deleted-node stack. Only called by dream().
 * Note - a deleted node is marked by having empty input and output lists
 */
 
void delete_node(net_type *net,int node)
{

  int lhs, rhs;
  
  /* printf("Deleting node...\n"); */
  lhs=net->data[node+LHS_IP]; rhs=net->data[node+RHS_IP];
  
  /* sanity check... we cant delete a node if it has any outputs*/
  if (net->data[node+NUM_LHS_OP]!=0)
  {
     printf("WOA!!! delete_node() something well wrong!!\n");
     exit(0);
  }
  if (net->data[node+NUM_RHS_OP]!=0)
  {
     printf("WOA!!! delete_node() something well wrong!!\n");
     exit(0);
  }
  /* Firstly, if the node is already deleted, forget it. */
  if (lhs==0 ) return;
  
  /* Just need to delete inputs, then add node to the deleted_node stack*/
  
  del_lhs_output(net,lhs,node);  del_rhs_output(net,rhs,node);
  net->data[node+LHS_IP]=0; net->data[node+RHS_IP]=0;
  net->data[node+LHS_OP]=0; net->data[node+RHS_OP]=0; 
  /* OK, lastly, put the net->node item into the deleted_node stack */
  /* MODIFICATION - IF the node is at the TOP of the nodespace (this is
     very likely, given that dream deleted nodes backwards), then decrease the
     nodespace instead. */
  if (node==net->nodespace)
    net->nodespace=net->nodespace-TOP;
  else
  {
    net->data[node+SPEC]=net->deleted_node;
    net->deleted_node=node;
  }
  net->actual_num_nodes--;
}
